# -*- coding: utf-8 -*-
import turbogears as tg
import calendar,time,tw.forms,tw.core.resources,genshi
from turbogears import controllers, expose, flash, validate, error_handler
from uzi.model import *
from turbogears import identity, redirect
from cherrypy import request

from tw.api import WidgetsList, CSSSource, JSSource, js_function
from tw.forms import *
from tw.forms.validators import *
from formencode.api import Invalid

# from uzi import json
# import logging
# log = logging.getLogger("uzi.controllers")

u=lambda t: unicode(t,'utf-8')

class TGSchema(Schema):
     filter_extra_fields = True
     allow_extra_fields = True

def make_edit_link(obj,edit_url):
    name=obj.__str__()
    if not obj.visible: name+=u('<font color="#ff0000"> (удалён%s)</font>'%{True:'',False:'о'}['d' in edit_url])
    return make_link(name,edit_url,{'id':obj.id},anchor='#edit')

def make_link(name, url, args={}, args2={}, anchor=''):
    args.update(args2)
    return genshi.XML(u"<a href='%s#%s'>%s</a>"%(tg.url(url,args),anchor,name))

class PerfomanceObject:
    def __init__(self,name,score=0): self._done,self.name,self.score=0,name,score
    def feed(self, key, val):
        try: self.__dict__[key]+=val
        except KeyError: self.__dict__[key]=val
    def __getattr__(self,k):
        if self._done: return '-'
        else: raise AttributeError
    def done(self):
        self._done=1
        if self.score<=1: return
        for k in self.__dict__.keys():
            try: int(k)
            except ValueError: continue
            val=getattr(self,k)
            setattr(self,k,'%s (%s)'%(val,val/self.score))

class FieldsSumMatch(FieldsMatch):
    """ Validates that value of the first field in the given sequence
        matches sum of values of other fields in sequence with the precision of 10^-4.
        If given less than two fields always raises an Exception.
        If given only two fields behaves as primitive version of FieldsMatch validator.
        """
    def validate_python(self, field_dict, state):
#        import pdb;pdb.set_trace()
        try:
            val,sum= float(field_dict[self.field_names[0]]),0.0
        except: raise Invalid(u('Не указана общая сумма!'),'',state,error_dict={self.field_names[0]:u('Введите значение.')})
        try:
            errors={}
            for name in self.field_names[1:]:
                try:
                    sum+=float(field_dict[name])
                    errors[name]=u('Сумма не совпадает.')
                except: pass
            errors[self.field_names[0]]=u('Проверочное значение %s.'%int(sum))
            diff=val-sum
            assert diff*diff<0.00000001
        except:
            raise Invalid(u('Сумма не совпадает: %s != %s'%(sum,val)),'',state,error_dict=errors)

class FirstNotGreaterOrBothEmpty(FieldsMatch):
    """ Validates that value first given field value is lesser or equal to second field value.
        Allows both fields to be empty
        """
    def validate_python(self, field_dict, state):
#        import pdb;pdb.set_trace()
        errors={}
        try: val1 = float(field_dict[self.field_names[0]])
        except: val1=0
        try: val2 = float(field_dict[self.field_names[1]])
        except: val2=0
        if val2 and not val1: errors[self.field_names[0]]=u('Значение слишком мало.')
        if val1>val2:
            errors.update({
                self.field_names[1]:u('Значение слишком мало.'),
                self.field_names[0]:u('Значение слишком велико.'),})
        if errors:
            raise Invalid('','',state,error_dict=errors)

class DoctorValidator(FirstNotGreaterOrBothEmpty):
    " Is it doctor id? "
#    messages = {'missingdoctor': u("Выберите специалиста"),}
    def validate_python(self, field_dict, state):
#        import pdb;pdb.set_trace()
        errors={}
        dids=[d.id for d in Doctor.query.filter(Doctor.visible==1).all()]
        for name in self.field_names:
            try: val=int(field_dict.get(name,''))
            except (ValueError, TypeError): val=-1
            if val not in dids:
                errors[name]=u("Выберите специалиста")
        if errors:
            raise Invalid('','',state,error_dict=errors)

class NewResultForm(TableForm):
    action='/enter_data_confirm'
    def __init__(self, **dt):
        HiddenField('id', self, validator = Int)
        RadioButtonList('doctor',self,
            label_text = u("Специалист"),
            #          VVV removed `` to allow reedit
            options = [(d.id,d) for d in Doctor.query.filter(Doctor.visible==1).all()],
            validator = Int,
            )
        CalendarDatePicker('date',self,label_text = u("Дата"),date_format='%Y-%m-%d',button_text=u('Изменить'),calendar_lang='ru',
                            validator=Regex("^20\d\d-(0\d|1[012])(-([012]\d|3[01])|)$"))
        chain1,chain2,chain3,chain4,chain5=[],[],[],[],[]
        for o in Organ.query.all():
            fname='organ_%i'%o.id
            TextField(fname, self, label_text = o.name, size=2, validator = Int)
            #               1=results; 2=patients; 3=res>pati; 4=Pres>Ppati; 5=PUE>Pres
            if   o.type==0: chain1.append(fname)
            elif o.type==1: chain1.insert(0,fname) ; chain3.append(fname)
            elif o.type==2: chain2.insert(0,fname)                        ; chain3.insert(0,fname)
            elif o.type==3: chain2.append(fname)
            elif o.type==4:                          chain5.append(fname) 
            elif o.type==5:                          chain4.append(fname) ; chain5.insert(0,fname)
#            elif o.type==5: chain1.append(fname)   ; chain4.append(fname) ; chain5.insert(0,fname)
            elif o.type==6: chain2.append(fname)   ;                        chain4.insert(0,fname)
#1>=2 #2=3+3+6 #5>=6 #4>=5 #1=sum(0)
        self.validator = TGSchema(chained_validators = [
        DoctorValidator('doctor'), FieldsSumMatch(*chain1),FieldsSumMatch(*chain2),
        FirstNotGreaterOrBothEmpty(*chain3),FirstNotGreaterOrBothEmpty(*chain4),FirstNotGreaterOrBothEmpty(*chain5),
        ], if_key_missing = 0, )
        SubmitButton('submit', self,  default = u('Ввести'))
        TableForm.__init__(self, **dt)

#new_result_form=NewResultForm(validator=TGSchema)

class ResultConfirmForm(TableForm):
    action='/enter_data_do'
    def __init__(self, **dt):
        HiddenField('id', self, validator = Int)
        TextField('doctorname',  self, label_text = u("Специалист"))
        HiddenField('doctor',self, validator = Int)
        TextField('date',    self, label_text = u("Дата"), size=15,validator=Regex("^20\d\d-(0\d|1[012])(-([012]\d|3[01])|)$"))
        for o in Organ.query.all():
            TextField('organ_%i'%o.id, self, label_text = o.name, size=2, validator = Int)
        SubmitButton('submit', self,  default = u('Подтвердить'))
        TableForm.__init__(self, **dt)
#result_confirm_form=ResultConfirmForm(validator=TGSchema)

class ResultsQueryForm(TableForm):
    class fields(WidgetsList):
        date1=CalendarDatePicker(label_text = u("Первый день периода"),date_format='%Y-%m-%d',button_text=u('Изменить'),calendar_lang='ru',
                default=time.strftime('%Y-%m-01'))
        date30=CalendarDatePicker(label_text = u("Последний день периода"),date_format='%Y-%m-%d',button_text=u('Изменить'),calendar_lang='ru',
                default=time.strftime('%Y-%m-')+`calendar.monthrange(*(time.localtime()[:2]))[1]`)
        doctors=MultipleSelectField(options=[(`doctor.id`,doctor) for doctor in Doctor.query.filter(Doctor.visible==1).all()],
                label_text = u("Отчёт по специалистам"),size=len(Doctor.query.filter(Doctor.visible==1).all()),
                default=[`doctor.id` for doctor in Doctor.query.filter(Doctor.visible==1).all()])
        submit=SubmitButton(default = u('Сгенерировать отчёт'))
        sumbit2=SubmitButton(default = u('Вариант для печати'), attrs={'name':'print'})
#results_query_form=ResultsQueryForm(validator=TGSchema)

class DoctorEditForm(TableForm):
    action='/doctor_edit'
    class fields(WidgetsList):
        id=HiddenField(validator = Int)
        label=Label(label_text = u("Добавить специалиста"))
        surname=TextField(label_text = u("Фамилия"), validator = UnicodeString(min=1))
        first_name=TextField(label_text = u("Имя"), validator = UnicodeString())
        last_name=TextField(label_text = u("Отчество"), validator = UnicodeString())
        order=TextField(label_text = u("Место в списке"), size=3, validator = Int(if_empty=100))
#        norm=TextField(label_text = u("Норма УЕ"), size=4, validator = Int(if_empty=0))
        datetype=RadioButtonList(
            label_text = u("Отчётность"),
            options = [(1,u("ежедневная")),(30, u("ежемесячная"))],
            validator = Int,
            )
        submit=SubmitButton(default = u('Добавить'))
        submit_del=SubmitButton(default = u('Убрать из отчёта'), attrs={'name':'delete'})
        submit_del2=SubmitButton(default = u('Удалить совсем'), attrs={'name':'erase'})
doctor_edit_form=DoctorEditForm(validator=TGSchema)

class OrganEditForm(TableForm):
    action='/organ_edit'
    class fields(WidgetsList):
        id=HiddenField(validator = Int)
        label=Label(label_text = u("Добавить пункт отчётности"))
        name=TextField(label_text = u("Наименование"), size=40, validator = UnicodeString(min=1))
        order=TextField(label_text = u("Место в списке"), size=3, validator = Int(if_empty=100))
        score=TextField(label_text = u("Количество УЕ"), size=3, validator = Number(if_empty=0))
        submit=SubmitButton(default = u('Добавить'))
        submit_del=SubmitButton(default = u('Убрать из отчёта'), attrs={'name':'delete'})
        submit_del2=SubmitButton(default = u('Удалить совсем'), attrs={'name':'erase'})
organ_edit_form=OrganEditForm(validator=TGSchema)

#==========================================================

class Root(controllers.RootController):

    @expose()
    def index(self):
        redirect("/enter_data_form")

    @expose(template="uzi.templates.login")
    def login(self, forward_url=None, *args, **kw):

        if forward_url:
            if isinstance(forward_url, list):
                forward_url = forward_url.pop(0)
            else:
                del request.params['forward_url']

        if not identity.current.anonymous and identity.was_login_attempted() \
                and not identity.get_identity_errors():
            redirect(tg.url(forward_url or '/', kw))

        if identity.was_login_attempted():
            msg = _("The credentials you supplied were not correct or "
                   "did not grant access to this resource.")
        elif identity.get_identity_errors():
            msg = _("You must provide your credentials before accessing "
                   "this resource.")
            if not forward_url:
                forward_url = request.path_info
        else:
            msg = _("Please log in.")
            if not forward_url:
                forward_url = request.headers.get("Referer", "/")

        # we do not set the response status here anymore since it
        # is now handled in the identity exception.
        return dict(logging_in=True, message=msg,
            forward_url=forward_url, previous_url=request.path_info,
            original_parameters=request.params)

    @expose()
    def logout(self):
        identity.current.logout()
        redirect("/")

    @expose(template="uzi.templates.enter_data_form")
    def enter_data_form(self, **dt):
        if dt.get('edit_id',None):
            res=USG_Result.get_by(id=int(dt['edit_id']))
            dt={'doctor':`res.doctor.id`,'date':res.date, 'id':res.id}
            for o in res.subresults: dt['organ_%i'%o.organ.id]=o.number
        return dict(result_form=NewResultForm(), data=dt)

    @expose(template="uzi.templates.enter_data_form")
    @validate(form=NewResultForm())
    @error_handler(enter_data_form)
    def enter_data_confirm(self, **dt):
#	import pdb;pdb.set_trace()
#        if dt.get('doctor', None):
        dt['doctorname']=Doctor.get_by(id=int(dt.get('doctor', None)))
        return dict(result_form=ResultConfirmForm(validator=TGSchema), data=dt)

    def date2datetime(self, date):
        import platform
        date=str(date)+'0001-01-01 23:59:59'[len(str(date)):]
        if platform.python_version()>='2.5':
            date=datetime(*[(int(date[k*3+2:k*3+4])+2000*(not k)) for k in range(6)])
        return date

    @expose()
    @validate(form=NewResultForm())
    @error_handler(enter_data_form)
    def enter_data_do(self, **dt):
        if dt.get('id', None) and not identity.in_group("admin"):
            dt['forward_url']='/enter_data_do'
            redirect(tg.url("/login",dt))
        date=dt.get('date')
        doctor=Doctor.get_by(id=int(dt.get('doctor', None)))
        if not date or not doctor: redirect(tg.url('/',dt))
        if dt.get('id', None): res=USG_Result.get_by(id=int(dt['id']))
        else: res=USG_Result()
        date=str(date)+'0001-01-01 23:59:59'[len(str(date)):]
        res.date=self.date2datetime(date)
        if doctor.datetype==30: date=date[:7]
        else: date=date[:10]
        res.doctor=doctor
#        import pdb;pdb.set_trace()
        for i in dt:
            if i[:6]=='organ_':
                for sr in res.subresults[:]:
                    if sr.organ.id==int(i[6:]):
                        if dt[i] and int(dt[i]): sr.number=int(dt[i])
                        else: sr.delete()
                        break
                else:
                    if dt[i] and int(dt[i]):
                       res.subresults.append(USG_SubResult(organ=Organ.get_by(id=int(i[6:])),number=int(dt[i])))
        flash(u('Данные специалита %s за %s %s')%(doctor,date,{True:u('добавлены'),False:u('исправлены')}[not dt.get('id', None)]))
        if dt.get('id', None): redirect(tg.url("/view_data",dict(doctors=doctor.id)))
        redirect("/")

    @expose(template="uzi.templates.print_report")
    @identity.require(identity.in_group("admin"))
    def print_report(self, **dt): return self.report_data(**dt)

    @expose(template="uzi.templates.report_data")
    @identity.require(identity.in_group("admin"))
    def preview_report(self, **dt): return self.report_data(**dt)

    def report_data(self, **dt):
        """ This is the main report page """
        if dt.has_key('print'):
            del dt['print']
            redirect(tg.url("/print_report",dt))
            return
        perf=PerfomanceObject('top')
        perf.total=PerfomanceObject(u('Фактич. выполнено'))
#        perf.norm=PerfomanceObject(u('Норма УЕ'))
        query=USG_Result.query

        if not dt.get('date1',None): dt['date1']=time.strftime('%Y-%m-01')
        if not dt.get('date30',None): dt['date30']=time.strftime('%Y-%m-')+`calendar.monthrange(*(time.localtime()[:2]))[1]`
        if dt.get('date1',None): query=query.filter(USG_Result.date>=self.date2datetime(dt.get('date1')))
        if dt.get('date30',None): query=query.filter(USG_Result.date<=self.date2datetime(dt.get('date30')+'0001-12-31 23:59:59'[len(dt.get('date30')):]))
        docs=[]
        for result in query.all():
            for subresult in result.subresults:
                if not result.doctor or not result.doctor.visible: continue
                if dt.get('doctors',None) and `result.doctor.id` not in dt['doctors']: continue
                oid,did=subresult.organ.id,result.doctor.id,
                if result.doctor not in docs: docs.append(result.doctor)
                try: row=getattr(perf,`oid`)
                except AttributeError:
                    row=PerfomanceObject(subresult.organ.name, subresult.organ.score)
                    setattr(perf,`oid`,row)
                val=subresult.organ.score*subresult.number
                if not val: val=subresult.number
                row.feed(`did`,val)
                if result.doctor.datetype<2:
                    row.feed('total1',val)
                row.feed('total30',val)
                if subresult.organ.score:
                    perf.total.feed(`did`,val)
                    if result.doctor.datetype<2:
                        perf.total.feed('total1',val)
                    perf.total.feed('total30',val)
        data=[]
        for oid in ['total']+[o.id for o in Organ.query.all()]:
            try: data.append(getattr(perf,str(oid)))
            except AttributeError: continue
            data[-1].done()

        docs1=[(d,`d.id`,d.order) for d in docs if d.datetype<2]
        docs1.sort(lambda a,b:cmp(a[2],b[2]))
        docs1=[(a,b) for a,b,c in docs1]
        docs30=[(d,`d.id`,d.order) for d in docs if d.datetype>2]
        docs30.sort(lambda a,b:cmp(a[2],b[2]))
        docs30=[(a,b) for a,b,c in docs30]
        fields=[(u('Наименование исследования'),'name')]
        if docs1: fields+=docs1+[(u('Всего'),'total1')]
        if docs30: fields+=docs30+[(u('Итого'),'total30'),]
        return dict(
            results_query_form=ResultsQueryForm(validator=TGSchema),
            report_form=DataGrid(fields=fields),
            results=data,
            query=dt,
            pagetitle=u('Отчёт за %s...%s')%(dt.get('date1',None),dt.get('date30',None)),
        )

    @expose(template="uzi.templates.db_edit")
    @identity.require(identity.in_group("admin"))
    def doctors(self, **dt):
        return self.view_objs(Doctor, dt.get('id',None), u('Специалисты'), [
            (u('Специалист'),lambda x: make_edit_link(x,'/doctors')),
            (u('Дней в отчёте'),'datetype'),
            (u('Место в списке'),'order'),],doctor_edit_form,)

    @expose(template="uzi.templates.db_edit")
    @identity.require(identity.in_group("admin"))
    def organs(self, **dt):
        return self.view_objs(Organ, dt.get('id',None), u('Поля отчёта'), [
            (u('Наименование'),lambda x: make_edit_link(x,'/organs'),),
            (u('УЕ'),'score'),
            (u('Место в списке'),'order'),], organ_edit_form)

    def view_objs(self, Obj, obj, pagetitle, fields, obj_edit_form):
        if obj: obj=Obj.get_by(id=obj)
        objs=Obj.query
        if Obj==Organ: objs=objs.filter(Organ.type==0)
        return dict(
            view_form=DataGrid(fields=fields),
            view_data=objs.all(),
            edit_form=obj_edit_form,
            edit_data=obj,
            pagetitle=pagetitle,
        )

    @expose()
    @validate(form=doctor_edit_form)
    @error_handler(doctors)
    @identity.require(identity.in_group("admin"))
    def doctor_edit(self, **dt):
        self.obj_edit(Doctor, dt, '/doctors')

    @expose()
    @validate(form=organ_edit_form)
    @error_handler(organs)
    @identity.require(identity.in_group("admin"))
    def organ_edit(self, **dt):
        self.obj_edit(Organ, dt, '/organs')

    def obj_edit(self, Obj, dt, url):
        obj=Obj.get_by(id=dt['id'])
        del dt['id']
        if getattr(obj,'type',None) and dt.get('erase',None): dt['delete']=1 ;redirect(tg.url(url))
        if obj and dt.get('delete',None):
            if not getattr(obj,'type',None):
                obj.visible=0
        elif obj and dt.get('erase',None):
            if not getattr(obj,'type',None):
                obj.delete()
        else:
            if not obj: obj=Obj(**dt)
            obj.visible=1
            for k,v in dt.iteritems(): setattr(obj,k,v)
        redirect(tg.url(url))

    @expose(template="uzi.templates.report_data")
    @identity.require(identity.in_group("admin"))
    def view_data(self, **dt):
#        import pdb;pdb.set_trace()
        perf=PerfomanceObject('top')
        query=USG_Result.query
        if not dt.get('date1',None): dt['date1']=time.strftime('%Y-%m-01')
        if not dt.get('date30',None): dt['date30']=time.strftime('%Y-%m-')+`calendar.monthrange(*(time.localtime()[:2]))[1]`
        if dt.get('date1',None): query=query.filter(USG_Result.date>=self.date2datetime(dt.get('date1')))
        if dt.get('date30',None): query=query.filter(USG_Result.date<=self.date2datetime(dt.get('date30')+'0001-12-31 23:59:59'[len(dt.get('date30')):]))
        docs=[]
        for result in query.all():
            if not result.doctor: continue
            if dt.get('doctors',None) and `result.doctor.id` not in dt['doctors']: continue
            docs.append((result.doctor, result.date, result.id))
            for subresult in result.subresults:
                if not result.doctor.visible: continue
                oid=subresult.organ.id
                try: row=getattr(perf,`oid`)
                except AttributeError:
                    row=PerfomanceObject(subresult.organ.name)
                    setattr(perf,`oid`,row)
                row.feed(str(result.id),subresult.number)

        data=[]
        for oid in [o.id for o in Organ.query.all()]:
            try: data.append(getattr(perf,`oid`))
            except AttributeError: continue
            data[-1].done()
        fields=[(u('Наименование исследования'),'name')]
        for doc,date,id in docs:
            title=make_link(u'%s %s'%(doc.__str__(),str(date)[:{1:10,30:7}[doc.datetype]]), "/enter_data_form", {'edit_id':id})
            fields.append((title,str(id)))
        return dict(
            results_query_form=ResultsQueryForm(validator=TGSchema),
            report_form=DataGrid(fields=fields),
            results=data,
            query=dt,
        )
