# -*- coding: utf-8 -*-
from datetime import datetime
import pkg_resources
pkg_resources.require("SQLAlchemy>=0.4.0")
pkg_resources.require("Elixir>=0.4.0")
# import the basic Elixir classes and functions for declaring the data model
# (see http://elixir.ematia.de/trac/wiki/TutorialDivingIn)
from elixir import Entity, Field, OneToOne, OneToMany, ManyToOne, ManyToMany
from elixir import options_defaults, using_options, setup_all
# import some datatypes for table columns from Elixir
# (see http://www.sqlalchemy.org/docs/04/types.html for more)
from elixir import String, Unicode, Integer, DateTime, Float, Boolean
from turbogears import identity

options_defaults['autosetup'] = True


# your data model

# class YourDataClass(Entity):
#     pass

class Doctor(Entity):
    using_options(order_by=['order', 'surname'])
    surname = Field(String(40))
    first_name = Field(String(40))
    last_name = Field(String(40))
    datetype = Field(Integer, default=1)
    visible = Field(Boolean, default=1)
    order = Field(Integer)
    norm = Field(Integer)
    def __repr__(self): return '<Doctor %s>' % (self.surname.encode('utf-8'))
    def __str__(self): return (u'%s %s.%s.'%(self.surname,self.first_name[:1],self.last_name[:1])).replace('  . .','')
    submit=unicode('Сохранить','utf-8')

class Organ(Entity):
    using_options(order_by=['order', 'name'])
    name = Field(String(40), unique=True)
    score = Field(Float)
    order = Field(Integer)
    visible = Field(Boolean, default=1)
    type = Field(Integer)
    def __repr__(self): return '<Organ %s (%i)>' % (self.name.encode('utf-8'), self.score)
    def __str__(self): return self.name
    submit=unicode('Сохранить','utf-8')

class USG_Result(Entity):
    using_options(order_by=['date'])
    date = Field(DateTime, nullable=False, default=datetime.now)
    doctor = ManyToOne('Doctor')
    subresults = OneToMany('USG_SubResult')
    def __repr__(self): return '<USG_Result %s %s %s>' % (self.doctor.surname.encode('utf-8'), `self.date`, `self.subresults`)
#    def __getattr__(self, k):
#        if k[:6]=='organ_':
#            sr=self.getsubresult(int(k[6:]))
#            if sr: return '%s (%s)'%(os[0].number*os[0].organ.score,os[0].number,)
#            else: return '-'
#        else:
#            raise AttributeError #eturn Entity.__getattr__(self, k)
#    def getsubresult(self, oid):
#        for sr in self.subresults:
#            if sr.organ.id==oid: return sr

class USG_SubResult(Entity):
    result = ManyToOne('USG_Result')
    organ = ManyToOne('Organ')
    number = Field(Integer)
    def __repr__(self): return '<USG_SubResult %s x %s>' % (self.organ.name.encode('utf-8'), self.number)
#    comment = Field(String(250))


# the identity model


class Visit(Entity):
    """
    A visit to your site
    """
    using_options(tablename='visit')

    visit_key = Field(String(40), primary_key=True)
    created = Field(DateTime, nullable=False, default=datetime.now)
    expiry = Field(DateTime)

    @classmethod
    def lookup_visit(cls, visit_key):
        return Visit.get(visit_key)


class VisitIdentity(Entity):
    """
    A Visit that is link to a User object
    """
    using_options(tablename='visit_identity')

    visit_key = Field(String(40), primary_key=True)
    user = ManyToOne('User', colname='user_id', use_alter=True)


class Group(Entity):
    """
    An ultra-simple group definition.
    """
    using_options(tablename='tg_group')

    group_id = Field(Integer, primary_key=True)
    group_name = Field(Unicode(16), unique=True)
    display_name = Field(Unicode(255))
    created = Field(DateTime, default=datetime.now)
    users = ManyToMany('User', tablename='user_group')
    permissions = ManyToMany('Permission', tablename='group_permission')


class User(Entity):
    """
    Reasonably basic User definition.
    Probably would want additional attributes.
    """
    using_options(tablename='tg_user')

    user_id = Field(Integer, primary_key=True)
    user_name = Field(Unicode(16), unique=True)
    email_address = Field(Unicode(255), unique=True)
    display_name = Field(Unicode(255))
    password = Field(Unicode(40))
    created = Field(DateTime, default=datetime.now)
    groups = ManyToMany('Group', tablename='user_group')

    @property
    def permissions(self):
        p = set()
        for g in self.groups:
            p |= set(g.permissions)
        return p


class Permission(Entity):
    """
    A relationship that determines what each Group can do
    """
    using_options(tablename='permission')

    permission_id = Field(Integer, primary_key=True)
    permission_name = Field(Unicode(16), unique=True)
    description = Field(Unicode(255))
    groups = ManyToMany('Group', tablename='group_permission')


# Set up all Elixir entities declared above

setup_all()
